# Copyright 2021 Juan L Gamella

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
"""

import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
import itertools


def argmin(array):
    """Return the index of the minimum element of an array.

    Examples
    --------
    >>> _argmin(np.array([1,-1,2,3]))
    (1,)
    >>> _argmin(np.array([[0,1],[-1,1]]))
    (1, 0)
    """
    return np.unravel_index(np.argmin(array), array.shape)


def argmax(array):
    """Return the index of the minimum element of an array.

    Examples
    --------
    >>> _argmin(np.array([1,-1,2,3]))
    (3,)
    >>> _argmin(np.array([[0,1],[2,1]]))
    (1, 0)
    """
    return np.unravel_index(np.argmax(array), array.shape)


def matrix_block(M, rows, cols):
    """
    Select a block of a matrix given by the row and column indices
    """
    return M[rows, :][:, cols]


def sampling_matrix(W):
    """Given the weighted adjacency matrix of a DAG, return
    the matrix A such that the DAG generates samples
      A @ diag(var)^1/2 @ Z + mu
    where Z is an isotropic normal, and var/mu are the variances/means
    of the noise variables of the graph.
    """
    p = len(W)
    return np.linalg.inv(np.eye(p) - W.T)


def all_but(k, p):
    """Return [0,...,p-1] without k"""
    k = np.atleast_1d(k)
    return [i for i in range(p) if i not in k]


def combinations(p, target, empty=True):
    """Return all possible subsets of the set {0...p-1} \\ {target}"""
    base = set(range(p)) - {target}
    sets = []
    for size in range(0 if empty else 1, p):
        sets += [set(s) for s in itertools.combinations(base, size)]
    return sets


def nonzero(A, tol=1e-12):
    """Return the indices of the nonzero (up to tol) elements in A"""
    return np.where(np.abs(A) > tol)[0]


def graph_info(i, W, interventions=None):
    """Returns the parents, children, parents of children and markov
    blanket of variable i in DAG W, using the graph structure
    """
    G = nx.from_numpy_array(W, create_using=nx.DiGraph)
    parents = set(G.predecessors(i))
    children = set(G.successors(i))
    parents_of_children = set()
    for child in children:
        parents_of_children.update(G.predecessors(child))
    if len(children) > 0:
        parents_of_children.remove(i)
    mb = parents.union(children, parents_of_children)
    return (parents, children, parents_of_children, mb)


def stable_blanket(i, W, interventions=set()):
    """Return the stable blanket using the graph structure"""
    G = nx.from_numpy_array(W, create_using=nx.DiGraph)
    parents = set(G.predecessors(i))
    children = set(G.successors(i))
    unstable_descendants = set()
    for j in interventions:
        if j in children:
            unstable_descendants.update({j})
            unstable_descendants.update(nx.algorithms.dag.descendants(G, j))
    stable_children = set.difference(children, unstable_descendants)
    parents_of_stable_children = set()
    for child in stable_children:
        parents_of_stable_children.update(G.predecessors(child))
    if len(stable_children) > 0:
        parents_of_stable_children.remove(i)
    sb = set.union(parents, stable_children, parents_of_stable_children)
    return sb


def descendants(i, W):
    """Return the descendants of a node using the graph structure"""
    G = nx.from_numpy_array(W, create_using=nx.DiGraph)
    return nx.algorithms.dag.descendants(G, i)


def ancestors(i, W):
    """Return the ancestors of a node using the graph structure"""
    G = nx.from_numpy_array(W, create_using=nx.DiGraph)
    return nx.algorithms.dag.ancestors(G, i)


def is_dag(A):
    """Checks wether the given adjacency matrix corresponds to a DAG.

    Parameters
    ----------
    A : np.array
        the adjacency matrix of the graph, where A[i,j] != 0 => i -> j.

    Returns
    -------
    is_dag : bool
        if the adjacency corresponds to a DAG

    """
    try:
        topological_ordering(A)
        return True
    except ValueError:
        return False


def topological_ordering(A):
    """Return a topological ordering for the DAG with adjacency matrix A,
    using Kahn's 1962 algorithm.

    Raises a ValueError exception if the given adjacency does not
    correspond to a DAG.

    Parameters
    ----------
    A : np.array
        The adjacency matrix of the graph, where A[i,j] != 0 => i -> j.

    Returns
    -------
    ordering : list of ints
        A topological ordering for the DAG.

    Raises
    ------
    ValueError
        If the given adjacency does not correspond to a DAG.

    """
    # Check that there are no undirected edges
    if (np.logical_and(A, A.T)).sum() > 0:
        raise ValueError("The given graph is not a DAG")
    # Run the algorithm from the 1962 paper "Topological sorting of
    # large networks" by AB Kahn
    A = A.copy()
    sinks = list(np.where(A.sum(axis=0) == 0)[0])
    ordering = []
    while len(sinks) > 0:
        i = sinks.pop()
        ordering.append(i)
        for j in ch(i, A):
            A[i, j] = 0
            if len(pa(j, A)) == 0:
                sinks.append(j)
    # If A still contains edges there is at least one cycle
    if A.sum() > 0:
        raise ValueError("The given graph is not a DAG")
    else:
        return ordering


def edge_weights(W):
    """Return the weights of all the edges of W in a dictionary, i.e. with
    keys (i,j) for values W[i,j] when W[i,j] != 0."""
    # TODO: Test
    fro, to = np.where(W != 0)
    edges = list(zip(fro, to))
    weights = [W[i, j] for i, j in edges]
    edge_weights = dict(zip(edges, weights))
    return edge_weights


def plot_graph(W, labels=None, weights=False, block=False):
    """Plot a graph with weight matrix W."""
    G = nx.from_numpy_array(W, create_using=nx.DiGraph)
    pos = nx.drawing.layout.shell_layout(G, scale=0.5)
    p = len(W)
    if labels is None:
        node_labels = dict(zip(np.arange(p), map(lambda i: "$X_{%d}$" % i, range(p))))
    else:
        node_labels = dict(zip(np.arange(p), labels))
    # Plot
    fig = plt.figure()
    params = {'node_color': 'white',
              'edgecolors': 'black',
              'node_size': 900,
              'linewidths': 1.5,
              'width': 1.5,
              'arrowsize': 20,
              'arrowstyle': '->',
              'min_target_margin': 10,
              'labels': node_labels}
    nx.draw(G, pos, **params)
    # Edge weights
    if weights:
        formatted = dict((e, "%0.3f" % w) for (e, w) in edge_weights(W).items())
        nx.draw_networkx_edge_labels(G, pos, formatted, font_color='red')
    fig.set_facecolor("white")
    plt.show(block=block)


def plot_matrix(A, ax=None, vmin=-3, vmax=3, formt="%0.2f", thresh=1e-16, block=False):
    """Plot a heatmap for the given matrix A.

    Parameters
    ----------
    A : numpy.ndarray
        The matrix to plot.
    ax : matplotlib.pyplot.axis, optional
        The axis to plot on; if None, create a new figure.
    vmin : float, default=-3
        The lower threshold for color saturation.
    vmax : float, default=3
        The upper threshold for color saturation.
    formt : string, default="%0.2f"
        The format with which to print the values of the matrix on top
        of the corresponding cell.
    thresh : float, default=1e-16
        Elements of the matrix which are lower than the threshold in
        absolute value are plotted as a zero (i.e. white, no text)
    """
    if ax is None:
        plt.figure()
        ax = plt.gca()
    ax.imshow(A, vmin=vmin, vmax=vmax, cmap='bwr')
    for i in range(len(A)):
        for j in range(len(A)):
            if A[i, j] != 0:
                ax.text(j, i, formt % A[i, j], ha='center', va='center')
    plt.show(block=block)


def allclose(A, B, rtol=1e-5, atol=1e-8):
    """Use np.allclose to compare, but relative tolerance is relative to
    the smallest element compared
    """
    return np.allclose(np.maximum(A, B), np.minimum(A, B), rtol, atol)

# --------------------------------------------------------------------
# Graph functions for PDAGS


def na(y, x, A):
    """Return all neighbors of y which are adjacent to x in A.

    Parameters
    ----------
    y : int
        the node's index
    x : int
        the node's index
    A : np.array
        the adjacency matrix of the graph, where A[i,j] != 0 => i -> j
        and A[i,j] != 0 & A[j,i] != 0 => i - j.

    Returns
    -------
    nodes : set of ints
        the resulting nodes

    """
    return neighbors(y, A) & adj(x, A)


def neighbors(i, A):
    """The neighbors of i in A, i.e. all nodes connected to i by an
    undirected edge.

    Parameters
    ----------
    i : int
        the node's index
    A : np.array
        the adjacency matrix of the graph, where A[i,j] != 0 => i -> j
        and A[i,j] != 0 & A[j,i] != 0 => i - j.

    Returns
    -------
    nodes : set of ints
        the neighbor nodes

    """
    return set(np.where(np.logical_and(A[i, :] != 0, A[:, i] != 0))[0])


def adj(i, A):
    """The adjacent nodes of i in A, i.e. all nodes connected by a
    directed or undirected edge.
    Parameters
    ----------
    i : int
        the node's index
    A : np.array
        the adjacency matrix of the graph, where A[i,j] != 0 => i -> j
        and A[i,j] != 0 & A[j,i] != 0 => i - j.

    Returns
    -------
    nodes : set of ints
        the adjacent nodes

    """
    return set(np.where(np.logical_or(A[i, :] != 0, A[:, i] != 0))[0])


def pa(i, A):
    """The parents of i in A.

    Parameters
    ----------
    i : int
        the node's index
    A : np.array
        the adjacency matrix of the graph, where A[i,j] != 0 => i -> j
        and A[i,j] != 0 & A[j,i] != 0 => i - j.

    Returns
    -------
    nodes : set of ints
        the parent nodes

    """
    return set(np.where(np.logical_and(A[:, i] != 0, A[i, :] == 0))[0])


def ch(i, A):
    """The children of i in A.

    Parameters
    ----------
    A : np.array
        the adjacency matrix of the graph, where A[i,j] != 0 => i -> j
        and A[i,j] != 0 & A[j,i] != 0 => i - j.

    Returns
    -------
    nodes : set of ints
        the children nodes

    """
    return set(np.where(np.logical_and(A[i, :] != 0, A[:, i] == 0))[0])


def same_normal(sample_a, sample_b, atol=5e-2, debug=False):
    """
    Test (crudely, by L1 dist. of means and covariances) if samples
    from two distributions come from the same Gaussian
    """
    mean_a, mean_b = np.mean(sample_a, axis=0), np.mean(sample_b, axis=0)
    cov_a, cov_b = np.cov(sample_a, rowvar=False), np.cov(sample_b, rowvar=False)
    print("MEANS\n%s\n%s\n\nCOVARIANCES\n%s\n%s" %
          (mean_a, mean_b, cov_a, cov_b)) if debug else None
    means = np.allclose(mean_a, mean_b, atol=atol)
    covariances = np.allclose(cov_a, cov_b, atol=atol)
    return means and covariances

# Example graphs


def eg1():
    W = np.array([[0, 1, -1, 0, 0],
                  [0, 0, 0, 1, 0],
                  [0, 0, 0, 1, 0],
                  [0, 0, 0, 0, 1],
                  [0, 0, 0, 0, 0]])
    markov_blankets = [[1, 2],
                       [0, 3, 2],
                       [0, 3, 1],
                       [1, 2, 4],
                       [3]]
    parents = [[],
               [0],
               [0],
               [1, 2],
               [3]]
    return W, parents, markov_blankets


def eg2():
    W = np.array([[0, 1, -1, 0, 0, 0],
                  [0, 0, 0, 1, 0, 0],
                  [0, 0, 0, 1, 0, 0],
                  [0, 0, 0, 0, 1, 0],
                  [0, 0, 0, 0, 0, 0],
                  [0, 0, 0, 0, 1, 0]])
    markov_blankets = [[1, 2],
                       [0, 3, 2],
                       [0, 3, 1],
                       [1, 2, 4, 5],
                       [3, 5],
                       [3, 4]]
    parents = [[],
               [0],
               [0],
               [1, 2],
               [3, 5],
               []]
    return W, parents, markov_blankets


def eg3():
    W = np.array([[0, 1, -1, 0, 0, 0, 0, 0],
                  [0, 0, 0, 1, 0, 0, 0, 0],
                  [0, 0, 0, 1, 0, 0, 0, 0],
                  [0, 0, 0, 0, 1, 0, 0, 0],
                  [0, 0, 0, 0, 0, 0, 0, 1],
                  [0, 0, 0, 0, 1, 0, 1, 0],
                  [0, 0, 0, 0, 0, 0, 0, 0],
                  [0, 0, 0, 0, 0, 0, 0, 0]])
    markov_blankets = [[1, 2],
                       [0, 3, 2],
                       [0, 3, 1],
                       [1, 2, 4, 5],
                       [3, 5, 7],
                       [3, 4, 6],
                       [5],
                       [4]]
    parents = [[],
               [0],
               [0],
               [1, 2],
               [3, 5],
               [],
               [5],
               [4]]
    return W, parents, markov_blankets


def eg4():
    W = np.array([[0, 0, 1, 0],
                  [0, 0, 1, 0],
                  [0, 0, 0, 1],
                  [0, 0, 0, 0]])
    markov_blankets = [[1, 2],
                       [0, 2],
                       [0, 1, 3],
                       [2]]
    parents = [[],
               [],
               [0, 1],
               [2]]
    return W, parents, markov_blankets


def eg5():
    W = np.array([[0., 1., 0., 0., 0., 0., 0., 0.],
                  [0., 0., 0., 0., 0., 0., 0., 0.],
                  [0., 1., 0., 1., 0., 1., 0., 0.],
                  [0., 1., 0., 0., 0., 1., 1., 0.],
                  [1., 0., 0., 0., 0., 0., 1., 0.],
                  [0., 0., 0., 0., 0., 0., 0., 0.],
                  [1., 0., 0., 0., 0., 0., 0., 0.],
                  [1., 0., 0., 0., 1., 1., 0., 0.]])
    ordering = np.array([7, 4, 2, 3, 6, 5, 0, 1])
    parents = [[4, 6, 7],
               [0, 2, 3],
               [],
               [2],
               [7],
               [2, 3, 7],
               [3, 4],
               []]
    children = [[1],
                [],
                [1, 3, 5],
                [1, 5, 6],
                [0, 6],
                [],
                [0],
                [0, 4, 5]]
    markov_blankets = [[4, 6, 7, 1, 2, 3],
                       [0, 2, 3],
                       [1, 3, 5, 0, 7],
                       [0, 1, 2, 4, 5, 6, 7],
                       [7, 0, 6, 7, 3],
                       [2, 3, 7],
                       [3, 4, 0, 7],
                       [0, 4, 5, 2, 3, 6]]
    return W, parents, markov_blankets


def eg6():
    W = np.array([[0, 0, 1, 0, 1],
                  [0, 0, 1, 0, 1],
                  [0, 0, 0, 1, 0],
                  [0, 0, 0, 0, 1],
                  [0, 0, 0, 0, 0]])
    parents = [[],
               [],
               [0, 1],
               [2],
               [0, 1, 3]]
    children = [[2],
                [2],
                [3],
                [4],
                []]
    markov_blankets = [[1, 2, 3, 4],
                       [0, 2, 3, 4],
                       [0, 1, 3],
                       [2, 4, 0, 1],
                       [0, 1, 3]]
    return W, parents, markov_blankets



# Copyright 2021 Juan L Gamella

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Main module containing the sempler.LGANM class, to represent and
sample from a linear Gaussian additive noise model.
"""

import numpy as np
import sempler.utils as utils
from sempler.normal_distribution import NormalDistribution

# ---------------------------------------------------------------------
# LGANM class


class LGANM:
    """Represents a linear model with Gaussian additive noise.

    Parameters
    ----------
    W : array_like
        Connectivity (weights) matrix representing a DAG.
    variances : numpy.ndarray or tuple
        The variances of the noise terms, or a tuple representing
        the lower/upper bounds to sample them from a uniform
        distribution.
    means : numpy.ndarray or tuple
        The means of the noise terms, or a tuple representing
        the lower/upper bounds to sample them from a uniform
        distribution.

    Raises
    ------
    ValueError
        If the given connectivity does not correspond to a DAG.

    Examples
    --------

    Constructing a linear Gaussian SCM.

    >>> import sempler
    >>> import numpy as np

    (1) Define the connectivity matrix:

    >>> W = np.array([[0, 0, 0, 0.1, 0],
    ...               [0, 0, 2.1, 0, 0],
    ...               [0, 0, 0, 3.2, 0],
    ...               [0, 0, 0, 0, 5.0],
    ...               [0, 0, 0, 0, 0  ]])

    (2a) With explicit means and variances:

    >>> means = np.array([0,1,2,3,4])
    >>> variances = np.array([1,1,1,1,1])
    >>> lganm = sempler.LGANM(W, means, variances)

    (2b) With randomly sampled means and variances:

    >>> lganm = sempler.LGANM(W, (0,1), (0,1))

    An exception is thrown when the connectivity matrix does not correspond to a DAG:

    >>> A = [[0,1,0],[0,0,1],[1,0,0]]
    >>> sempler.LGANM(A, (0,0), (1,1))
    Traceback (most recent call last):
      ...
    ValueError: The given graph is not a DAG.

    Attributes
    ----------
    W : array_like
        Connectivity (weights) matrix representing a DAG.
    variances : numpy.ndarray
        The variances of the noise terms.
    means : numpy.ndarray
        The means of the noise terms.
    p : int
        The number of variables (size) of the SCM.

    """

    def __init__(self, W, means, variances):
        # Set connectivity matrix
        W = np.atleast_2d(W)
        if not utils.is_dag(W):
            raise ValueError("The given graph is not a DAG.")
        self.W = W.copy()
        self.p = len(W)

        # Set variances
        if isinstance(variances, tuple) and len(variances) == 2:
            self.variances = np.random.uniform(variances[0], variances[1], size=self.p)
        elif type(variances) == np.ndarray and len(variances) == self.p:
            self.variances = variances.copy()
        else:
            raise ValueError(
                "Unexpected value for variances. Expected a two-element tuple or numpy.ndarray of length p.")
        # Set means
        if isinstance(means, tuple) and len(means) == 2:
            self.means = np.random.uniform(means[0], means[1], size=self.p)
        elif type(means) == np.ndarray and len(means) == self.p:
            self.means = means.copy()
        else:
            raise ValueError(
                "Unexpected value for means. Expected a two-element tuple or numpy.ndarray of length p.")

    def sample(self, n=100, population=False, do_interventions={}, shift_interventions={}, noise_interventions={}, random_state=None):
        """Generates n observations from the linear Gaussian SCM, under the
        given do, shift or noise interventions. If none are given,
        sample from the observational distribution.

        Parameters
        ----------
        n : int, optional
            The size of the sample (i.e. number of
            observations). Defaults to 100.
        population : bool, optional
            If True, the function returns a symbolic normal
            distribution instead of samples (see
            sempler.NormalDistribution). Defaults to False.
        do_interventions : dict, optional
            A dictionary where keys correspond to the intervened
            variables, and the values are tuples representing the new
            mean/variance of the intervened variable, e.g. {1: (1,2)}.
        shift_interventions : dict, optional
            A dictionary where keys correspond to the intervened
            variables, and the values are tuples representing the
            mean/variance of the noise which is added to the
            intervened variables, e.g. {1: (1,2)}.
        noise_interventions : dict, optional
            A dictionary where keys correspond to the intervened
            variables, and the values are tuples representing the
            mean/variance of the new noise, e.g. {1: (1,2)}.
        random_state: int, optional
            To set the random state for reproducibility. Succesive
            calls with the same random state will return the same
            sample.

        Returns
        -------
        numpy.ndarray or sempler.NormalDistribution
            An array containing the sample, where each column
            corresponds to a variable; or, if population=True, a
            symbolic normal distribution (see
            sempler.NormalDistribution).

        Examples
        --------

        Sampling the observational environment in the "population setting"

        >>> distribution = lganm.sample(population = True)

        Sampling under a shift intervention on variable 1 with standard gaussian noise

        >>> samples = lganm.sample(100, shift_interventions = {1: (0,1)})

        Sampling under a noise intervention on variable 0 and a do intervention on variable 2:

        >>> samples = lganm.sample(100,
        ...                       noise_interventions = {0: (1,2)},
        ...                       do_interventions = {2 : (3,4)})

        Interventions can also be deterministic, i.e. setting a variable/noise term to a fixed value:

        >>> samples = lganm.sample(5, do_interventions = {2 : (99,0)})
        >>> samples[:,2]
        array([99., 99., 99., 99., 99.])

        """
        # Must copy as they can be changed by interventions, but we
        # still want to keep the observational SEM
        W = self.W.copy()
        variances = self.variances.copy()
        means = self.means.copy()

        # Perform shift interventions
        if shift_interventions:
            shift_interventions = _parse_interventions(shift_interventions)
            targets = shift_interventions[:, 0].astype(int)
            means[targets] += shift_interventions[:, 1]
            variances[targets] += shift_interventions[:, 2]

        # Perform noise interventions. Note that they take preference
        # i.e. "override" shift interventions
        if noise_interventions:
            noise_interventions = _parse_interventions(noise_interventions)
            targets = noise_interventions[:, 0].astype(int)
            means[targets] = noise_interventions[:, 1]
            variances[targets] = noise_interventions[:, 2]

        # Perform do interventions. Note that they take preference
        # i.e. "override" shift and noise interventions
        if do_interventions:
            do_interventions = _parse_interventions(do_interventions)
            targets = do_interventions[:, 0].astype(int)
            means[targets] = do_interventions[:, 1]
            variances[targets] = do_interventions[:, 2]
            W[:, targets] = 0

        # Sampling by building the joint distribution
        A = np.linalg.inv(np.eye(self.p) - W.T)
        mean = A @ means
        covariance = A @ np.diag(variances) @ A.T
        distribution = NormalDistribution(mean, covariance)
        if not population:
            return distribution.sample(n, random_state=random_state)
        else:
            return distribution


def _parse_interventions(interventions_dict):
    """Used internally by LGANM.sample. Transforms the interventions from
    a dictionary to an array"""
    interventions = []
    for (target, params) in interventions_dict.items():
        # Mean and variance provided
        if type(params) == tuple and len(params) == 2:
            interventions.append([target, params[0], params[1]])
        # Only mean provided, assume we're setting the variable to a deterministic value
        elif type(params) in [float, int]:
            interventions.append([target, params, 0])
        else:
            raise ValueError("Wrongly specified intervention")
    return np.array(interventions)


# To run the LGANM.sample doctests
if __name__ == '__main__':
    import doctest
    # Build LGANM
    W = np.array([[0, 0, 0, 0.1, 0],
                  [0, 0, 2.1, 0, 0],
                  [0, 0, 0, 3.2, 0],
                  [0, 0, 0, 0, 5.0],
                  [0, 0, 0, 0, 0]])
    lganm = LGANM(W, (0, 1), (0, 1))
    doctest.testmod(extraglobs={'lganm': lganm}, verbose=True)



# Copyright 2021 Juan L Gamella

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""Module containing the sempler.NormalDistribution class, which
represents a normal distribution (mean + covariance) and allows for
marginalization, conditioning and regression in the population setting.
"""

import numpy as np
import sempler.utils as utils
import warnings

# ---------------------------------------------------------------------
# NormalDistribution class


class NormalDistribution():
    """Symbolic representation of a normal distribution.

    Parameters
    ----------
    mean : array_like
        The marginal means of the variables.
    covariance : array_like
        The covariance matrix of the distribution.
    check_valid : {'ignore', 'warn', 'raise'}, optional
        Behaviour when the provided covariance matrix is not positive
        definite. Note that semi-definiteness is enough for sampling,
        but not for the conditioning and regression operations.

    Raises
    ------
    ValueError
        If `check_valid='raise'` and the provided matrix is not
        positive semidefinite; *or* if the sizes of the mean vector
        and covariance matrices do not match.

    Examples
    --------

    >>> import sempler
    >>> import numpy as np

    Defining a univariate standard normal distribution:

    >>> distribution = sempler.NormalDistribution(0, 1)

    Defining an isotropic normal with zero mean vector:

    >>> covariance = np.array([[1,0],[0,1]])
    >>> means = np.array([0,0])
    >>> distribution = sempler.NormalDistribution(means, covariance)

    An error is raised if the covariance matrix is not positive
    definite and we set `check_valid='raise'`:

    >>> sempler.NormalDistribution([0,0], [[1,0],[1,1]], check_valid='raise')
    Traceback (most recent call last):
      ...
    ValueError: Covariance matrix is not positive definite.

    >>> sempler.NormalDistribution([0,0], [[1,1],[1,1]], check_valid='raise')
    Traceback (most recent call last):
      ...
    ValueError: Covariance matrix is not positive definite.

    Or if the size of the mean vector and covariance matrix do not
    match:

    >>> sempler.NormalDistribution([0], [[1,0],[0,1]])
    Traceback (most recent call last):
      ...
    ValueError: Mismatch in the size of mean vector and covariance matrix.


    Attributes
    ----------
    mean : numpy.ndarray
        The marginal means of the variables.
    covariance : numpy.ndarray
        The covariance matrix of the distribution.
    p : int
        The number of variables.

    """

    def __init__(self, mean, covariance, check_valid='ignore'):
        mean = np.atleast_1d(mean)
        covariance = np.atleast_2d(covariance)
        # Check positive semidefiniteness
        if check_valid != 'ignore':
            try:
                np.linalg.cholesky(covariance)
            except np.linalg.LinAlgError:
                msg = "Covariance matrix is not positive definite."
                raise ValueError(msg) if check_valid == 'raise' else warnings.warn(msg)
        # Define other parameters
        if len(mean) != len(covariance):
            raise ValueError("Mismatch in the size of mean vector and covariance matrix.")
        self.p = len(mean)
        self.mean = mean.copy()
        self.covariance = covariance.copy()

    def __str__(self):
        return "mean:\n" + str(self.mean) + "\ncovariance:\n" + str(self.covariance)

    def sample(self, n, random_state=None):
        """Generate a sample from the distribution.

        Parameters
        ----------
        n : int
            The size of the sample (i.e. number of observations).
        random_state : int,optional
            To set the random state for reproducibility.

        Returns
        -------
        numpy.ndarray
            An array containing the sample, where each column
            corresponds to a variable.

        Example
        -------

        >>> distribution.sample(5, random_state = 42)
        array([[0.754143  , 1.67465613, 2.45882157, 2.4718102 , 3.8108411 ],
               [2.04520905, 1.21383225, 3.29393672, 4.46230358, 5.79975029],
               [0.99209988, 2.50969225, 4.41981256, 5.3086489 , 5.73704626],
               [0.76152114, 2.94130318, 4.45257529, 5.02206465, 6.25225282],
               [0.71256027, 1.22445591, 1.33646622, 2.19050731, 0.07261   ]])

        """
        np.random.seed(random_state) if random_state is not None else None
        return np.random.multivariate_normal(self.mean, self.covariance, size=n)

    def marginal(self, X):
        """Return the marginal distribution of some variables.

        Parameters
        ----------
        X : array_like
            The indices of the variables. Note that the indices in the
            new distribution are dependent on the order given in X,
            e.g. `marginal([0,1])` and `marginal([1,0])` yield
            different (permuted) distributions.

        Returns
        -------
        marginal : sempler.NormalDistribution
            The marginal distribution.

        Examples
        --------

        >>> marginal = distribution.marginal(0)

        >>> marginal = distribution.marginal([0,1])

        """
        # Parse params
        X = np.atleast_1d(X)
        # Compute marginal mean/variance
        mean = self.mean[X]
        covariance = utils.matrix_block(self.covariance, X, X)
        return NormalDistribution(mean, covariance)

    def conditional(self, Y, X, x):
        """Return the conditional distribution of some variables given some
        others' values.

        Parameters
        ----------
        Y : array_like, list of ints or np.array
            The indices of the conditioned variables. Note that the indices in the
            new distribution are dependent on the order given in Y,
            e.g. `marginal([0,1])` and `marginal([1,0])` yield
            different (permuted) distributions.
        X : array_like, list of ints or np.array
            The indices of the variables to condition on.
        x : array_like
            The values of the conditioning variables.

        Raises
        ------
        ValueError
            If the size of `X` and `x` do not match, or if `X` and `Y` are not disjoint.

        Returns
        -------
        distribution : sempler.NormalDistribution
            the conditional distribution.

        Examples
        --------

        Conditioning a single variable:

        >>> conditional = distribution.conditional(0, 1, .1)

        Conditioning several variables:

        >>> conditional = distribution.conditional([0,1], [2,3], [.2, .3])

        An exception is raised if the size of `X` and `x` do not match:

        >>> distribution.conditional(0, [1], [.1, .2])
        Traceback (most recent call last):
        ...
        ValueError: Mismatch in the size of X and x.

        or if `Y` and `X` are not disjoint:

        >>> distribution.conditional([0], [0,1], [0, .1])
        Traceback (most recent call last):
        ...
        ValueError: X and Y are not disjoint.

        """
        # Parse params
        Y = np.atleast_1d(Y)
        X = np.atleast_1d(X)
        x = np.atleast_1d(x)
        # Check exceptions
        if len(X) != len(x):
            raise ValueError("Mismatch in the size of X and x.")
        if len(set(Y) & set(X)) > 0:
            raise ValueError("X and Y are not disjoint.")
        # Conditioning on nothing = marginalizing
        if len(X) == 0:
            return self.marginal(Y)
        # See https://en.wikipedia.org/wiki/Multivariate_normal_distribution#Conditional_distributions
        cov_y = utils.matrix_block(self.covariance, Y, Y)
        cov_x = utils.matrix_block(self.covariance, X, X)
        cov_yx = utils.matrix_block(self.covariance, Y, X)
        cov_xy = utils.matrix_block(self.covariance, X, Y)
        mean_y = self.mean[Y]
        mean_x = self.mean[X]
        mean = mean_y + cov_yx @ np.linalg.inv(cov_x) @ (x - mean_x)
        covariance = cov_y - cov_yx @ np.linalg.inv(cov_x) @ cov_xy
        return NormalDistribution(mean, covariance)

    def regress(self, y, Xs):
        """Compute the population MLE of the regression coefficients and
        intercept from regressing a variable on a subset of others.

        Parameters
        ----------
        y : int
            The index of the response/predicted variable.
        Xs : array_like
            The indices of the predictor/explanatory variables.

        Returns
        -------
        coefs : numpy.ndarray
            A p-sized array containing the estimated coefficients in
            the indices in Xs and 0s elsewhere.
        intercept : float
            The estimated intercept.

        Example
        -------
        >>> distribution.regress(0,[1,2])
        (array([ 0.        , -0.33333333,  0.33333333,  0.        ,  0.        ]), 0.6666666666666666)

        Regressing a variable on itself:

        >>> distribution.regress(0,0)
        (array([1., 0., 0., 0., 0.]), 0.0)

        """
        coefs = np.zeros(self.p)
        # If predictors are given, perform regression, otherwise just fit
        # intercept
        Xs = np.atleast_1d(Xs)
        if len(Xs) > 0:
            cov_y_xs = self.covariance[y, Xs]  # utils.matrix_block(self.covariance, y, Xs)
            cov_xs = self.covariance[:, Xs][Xs, :]  # utils.matrix_block(self.covariance, Xs, Xs)
            coefs[Xs] = np.linalg.solve(cov_xs, cov_y_xs)
        intercept = self.mean[y] - coefs @ self.mean
        return (coefs, intercept)

    def mse(self, y, Xs):
        """Compute the population (i.e. expected) mean squared error resulting
        from regressing a variable on a subset of others.

        The regression coefficients/intercept are the MLE computed in
        :func:`~sempler.NormalDistribution.regress`.

        Parameters
        ----------
        y : int
            The index of the response/predicted variable.
        Xs : int, list of ints or np.array
            The indices of the predictor/explanatory variables.

        Returns
        -------
        mse : float
           the expected mean squared error.

        Example
        -------

        >>> distribution.mse(0, [1,2])
        0.24666666666666665

        Regressing a variable on itself yields a zero error:

        >>> distribution.mse(0,0)
        0.0

        """
        var_y = self.covariance[y, y]
        # Compute regression coefficients when regressing on Xs
        (coefs_xs, _) = self.regress(y, Xs)
        # Covariance matrix
        cov = self.covariance
        # Computing the MSE
        mse = var_y + coefs_xs @ cov @ coefs_xs.T - 2 * cov[y, :] @ coefs_xs.T
        return mse

    def equal(self, dist, rtol=1e-5, atol=1e-8):
        """Check if this distribution is equal to another
        sempler.NormalDistribution, up to a tolerance.

        Parameters
        ----------
        dist : sempler.NormalDistribution
            The distribution to compare with.
        rtol : float, optional
            The relative tolerance in the elements of the mean
            and covariance. Default is 1e-5.
        atol : float, optional
            The absolute tolerance in the elements of the mean
            and covariance. Default is 1e-8.

        Returns
        -------
        equal : bool
            If the two distributions have the same mean/covariance up
            to tolerance tol.

        Raises
        ------
        TypeError
           If `dist` is not a sempler.NormalDistribution.

        Example
        -------

        >>> import sempler
        >>> dist1 = sempler.NormalDistribution(0, 1)
        >>> sempler.NormalDistribution(0.5, 1).equal(dist1, atol=0.5)
        True
        >>> sempler.NormalDistribution(0.5, 1).equal(dist1)
        False

        An exception is raised if we attempt to compare with anything else than a sempler.NormalDistribution:

        >>> sempler.NormalDistribution(0,1).equal(1)
        Traceback (most recent call last):
        ...
        TypeError: Unexpected type for "dist".

        """
        # Check type
        if type(dist) != NormalDistribution:
            raise TypeError('Unexpected type for "dist".')
        equal = np.allclose(self.mean, dist.mean, atol=atol, rtol=rtol) and np.allclose(
            self.covariance, dist.covariance, atol=atol, rtol=rtol)
        return equal


# To run the method's doctests
if __name__ == '__main__':
    import doctest
    # Build LGANM
    covariance = np.array([[0.37, 0., 0.37, 0.37, 0.75],
                           [0., 0.95, 0.95, 0.95, 1.9],
                           [0.37, 0.95, 2.06, 2.06, 4.11],
                           [0.37, 0.95, 2.06, 2.66, 4.71],
                           [0.75, 1.9, 4.11, 4.71, 9.48]])
    distribution = NormalDistribution([1, 2, 3, 4, 5], covariance, check_valid='raise')
    doctest.testmod(extraglobs={'distribution': distribution}, verbose=True)



# Copyright 2021 Juan L Gamella

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# ---------------------------------------------------------------------
# Unit testing for module utils

import unittest
import numpy as np

import sempler.generators
import sempler.utils as utils

# Tests for the DAG generation


class GeneratorTests(unittest.TestCase):
    def test_avg_degree(self):
        p = 1000
        for k in range(1, 5):
            W = sempler.generators.dag_avg_deg(p, k, 1, 2)
            av_deg = np.sum(W > 0) * 2 / p
            self.assertEqual(len(W), p)
            self.assertTrue(av_deg - k < 0.5)
            self.assertTrue(utils.is_dag(W))

    def test_disconnected_graph(self):
        W = sempler.generators.dag_avg_deg(10, 0, 1, 1)
        self.assertEqual(np.sum(W), 0)

    def test_full_dag(self):
        for p in range(10):
            W = sempler.generators.dag_full(p)
            self.assertEqual(p * (p - 1) / 2, W.sum())

    def test_intervention_targets(self):
        possible_targets = set(range(10))
        # Test random-sized interventions
        interventions = sempler.generators.intervention_targets(10, 100, (0, 3))
        for intervention in interventions:
            self.assertLessEqual(len(intervention), 3)
            self.assertGreaterEqual(len(intervention), 0)
            self.assertEqual(len(intervention), len(set(intervention) & possible_targets))

        # Test empty-sized interventions
        interventions = sempler.generators.intervention_targets(10, 100, (0, 0))
        for intervention in interventions:
            self.assertEqual(len(intervention), 0)
        # Test empty-sized interventions (without replacement)
        interventions = sempler.generators.intervention_targets(10, 100, (0, 0), replace=False)
        for intervention in interventions:
            self.assertEqual(len(intervention), 0)

        # Test single-target interventions
        interventions = sempler.generators.intervention_targets(10, 100, 1)
        for intervention in interventions:
            self.assertEqual(len(intervention), 1)
            self.assertEqual(len(intervention), len(set(intervention) & possible_targets))

        # Test single-target interventions (without replacement)
        interventions = sempler.generators.intervention_targets(10, 10, 1, replace=False)
        union = set()
        for intervention in interventions:
            self.assertTrue(len(union & set(intervention)) == 0)
            union |= set(intervention)
            self.assertEqual(len(intervention), 1)
            self.assertEqual(len(intervention), len(set(intervention) & possible_targets))
        self.assertEqual(union, set(range(10)))

    def test_intervention_targets_without_replacement(self):
        # Check that targets are not repeated (random intervention sizes)
        for i in range(100):
            interventions = sempler.generators.intervention_targets(
                20, 5, (0, 4), replace=False, random_state=i)
            union = set()
            for intervention in interventions:
                self.assertTrue(len(union & set(intervention)) == 0)
                union |= set(intervention)

        # Check that targets are not repeated (fixed intervention sizes)
        for i in range(100):
            interventions = sempler.generators.intervention_targets(
                20, 5, 4, replace=False, random_state=i)
            union = set()
            for intervention in interventions:
                self.assertTrue(len(union & set(intervention)) == 0)
                union |= set(intervention)



# Copyright 2021 Juan L Gamella

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""This module contains functions to be used in SCM assignments. They
receive n observations of k variables in a nxk array and must either a
scalar or a nx1 array"""


def null(*args):
    return 0



# Copyright 2021 Juan L Gamella

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#---------------------------------------------------------------------
# Unit testing for module utils

import unittest
import numpy as np

import sempler
import sempler.generators

# Tested functions
import sempler.utils as utils

class UtilsTests(unittest.TestCase):
    def test_combinations(self):
        for p in range(2, 10):
            target = np.random.choice(p)
            combinations = utils.combinations(p, target)
            self.assertEqual(int(2**(p-1)), len(combinations))
            [self.assertTrue(target not in s) for s in combinations]

    def test_is_dag_1(self):
        # Should be correct
        A = np.array([[0, 0, 1, 0, 0],
                      [0, 0, 1, 0, 0],
                      [0, 0, 0, 1, 1],
                      [0, 0, 0, 0, 1],
                      [0, 0, 0, 0, 0]])
        self.assertTrue(utils.is_dag(A))        

    def test_is_dag_2(self):
        # DAG with a cycle
        A = np.array([[0, 0, 1, 0, 0],
                      [0, 0, 1, 0, 0],
                      [0, 0, 0, 1, 1],
                      [1, 0, 0, 0, 1],
                      [0, 0, 0, 0, 0]])
        self.assertFalse(utils.is_dag(A))

    def test_is_dag_3(self):
        # PDAG
        A = np.array([[0, 0, 1, 0, 0],
                      [0, 0, 1, 0, 0],
                      [0, 0, 0, 1, 1],
                      [0, 0, 0, 0, 1],
                      [0, 0, 0, 1, 0]])
        self.assertFalse(utils.is_dag(A))

    def test_topological_sort_1(self):
        A = np.array([[0, 0, 0, 0, 0],
                      [0, 0, 0, 0, 0],
                      [1, 1, 0, 0, 0],
                      [0, 0, 1, 0, 0],
                      [0, 0, 1, 1, 0]]).T
        order = utils.topological_ordering(A)
        possible_orders = [[0,1,2,3,4], [1, 0, 2, 3, 4]]
        self.assertIn(order, possible_orders)
        
    def test_topological_sort_2(self):
        A = np.array([[0, 0, 0, 1, 0],
                      [0, 0, 0, 0, 0],
                      [1, 1, 0, 0, 0],
                      [0, 0, 1, 0, 0],
                      [0, 0, 1, 1, 0]]).T
        try:
            utils.topological_ordering(A)
            self.fail()
        except:
            pass

    def test_topological_sort_3(self):
        G = 100
        p = 30
        for i in range(G):
            A = sempler.generators.dag_avg_deg(p,3,1,1)
            ordering = utils.topological_ordering(A)
            fro,to = np.where(A != 0)
            # Test that the ordering is correct, i.e. for every edge x
            # -> y in the graph, x appears before in the ordering
            for (x,y) in zip(fro,to):
                pos_x = np.where(np.array(ordering) == x)[0][0]
                pos_y = np.where(np.array(ordering) == y)[0][0]
                self.assertLess(pos_x, pos_y)
        print("Checked topological sorting for %d DAGs" % (i+1))
            
    def test_matrix_block(self):
        M = np.array([[11, 12, 13, 14],
                      [21, 22, 23, 24],
                      [31, 32, 33, 34],
                      [41, 42, 43, 44]])
        # Tests
        tests = [(range(4), range(4), M),
                 ([1,2], [3], np.array([[24, 34]]).T),
                 (range(4), [1], M[:,[1]]),
                 ([2], range(4), M[[2],:]),
                 ([0,1], [0,1], np.array([[11, 12], [21, 22]])),
                 ([0,1], [1,3], np.array([[12,14], [22, 24]])),
                 # Test order of indices is also respected
                 (range(3,-1,-1), range(3,-1,-1), M[::-1,::-1]),
                 (range(3,-1,-1), range(4), M[::-1,:]),
                 (range(3,-1,-1), [1], M[::-1,[1]]),
                 ([2], range(3,-1,-1), M[[2],::-1]),
                 ([1,0], [0,1], np.array([[21, 22], [11, 12]])),
                 ([0,1], [3,1], np.array([[14,12], [24, 22]]))]
        for test in tests:
            (A, B, truth) = test
            #print(A, B, truth, matrix_block(M, A, B))
            self.assertTrue((utils.matrix_block(M, A, B) == truth).all())

    def test_sampling_matrix(self):
        W = np.array([[0,0,1,0],
                      [0,0,1,0],
                      [0,0,0,1],
                      [0,0,0,0]])
        truth = np.array([[1,0,0,0],
                          [0,1,0,0],
                          [1,1,1,0],
                          [1,1,1,1]])
        self.assertTrue((truth == utils.sampling_matrix(W)).all())

    def test_nonzero(self):
        tol = 1e-12
        A = np.array([0, 1, tol, -1, tol/2, -tol*2])
        self.assertTrue((np.array([1,3,5]) == utils.nonzero(A)).all())

    def test_all_but(self):
        self.assertTrue([0, 1, 3, 4] == utils.all_but(2,5))
        self.assertTrue([0, 3, 4] == utils.all_but([1,2],5))

    def test_graph_info_1(self):
        W = np.array([[0., 1., 0., 0., 0., 0., 0., 0.],
                      [0., 0., 0., 0., 0., 0., 0., 0.],
                      [0., 1., 0., 1., 0., 1., 0., 0.],
                      [0., 1., 0., 0., 0., 1., 1., 0.],
                      [1., 0., 0., 0., 0., 0., 1., 0.],
                      [0., 0., 0., 0., 0., 0., 0., 0.],
                      [1., 0., 0., 0., 0., 0., 0., 0.],
                      [1., 0., 0., 0., 1., 1., 0., 0.]])
        true_parents = [{4,6,7},
                        {0,2,3},
                        set(),
                        {2},
                        {7},
                        {2,3,7},
                        {3,4},
                        set()]
        true_children = [{1},
                         set(),
                         {1,3,5},
                         {1,5,6},
                         {0,6},
                         set(),
                         {0},
                         {0,4,5}]
        true_poc = [{2,3},
                    set(),
                    {0,3,7},
                    {0,2,4,7},
                    {3,6,7},
                    set(),
                    {4,7},
                    {2,3,4,6}]
        true_mb = [{4,6,7,1,2,3},
                   {0,2,3},
                   {1,3,5,0,7},
                   {0,1,2,4,5,6,7},
                   {7,0,6,7,3},
                   {2,3,7},
                   {3,4,0,7},
                   {0,4,5,2,3,6}]
        for i in range(len(W)):
            #print("Testing info for node %d" %i)
            (parents, children, poc, mb) = utils.graph_info(i, W)
            #print(parents, children, poc, mb)
            self.assertEqual(parents, true_parents[i])
            self.assertEqual(children, true_children[i])
            self.assertEqual(poc, true_poc[i])
            self.assertEqual(mb, true_mb[i])

    def test_graph_info_2(self):
        graphs = [utils.eg1(), utils.eg2(), utils.eg3(), utils.eg4(), utils.eg5(), utils.eg6()]
        for k,graph in enumerate(graphs):
            (W, true_parents, true_mb) = graph
            for i in range(len(W)):
                #print("%d Testing info for node %d" % (k+1,i))
                (parents, children, poc, mb) = utils.graph_info(i, W)
                #print(parents, children, poc, mb)
                self.assertEqual(parents, set(true_parents[i]))
                self.assertEqual(mb, set(true_mb[i]))

    def test_stable_blanket(self):
        W, _, markov_blanket = utils.eg5()
        target = 3
        intervened_variables = [set(), {2}, {7}, {6}, {5}]
        truth = [markov_blanket,
                 markov_blanket,
                 markov_blanket,
                 {2, 5, 7},
                 {2}]
        interventions = set()
        for i, var in enumerate(intervened_variables):
            interventions.update(var)
            self.assertTrue(truth[i], utils.stable_blanket(target, W, interventions))



# Copyright 2021 Juan L Gamella

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
This module contains functions to generate random graphs, which can then be used together with :class:`sempler.ANM` or :class:`sempler.LGANM` to produce random SCMs.
"""

import numpy as np

# ---------------------------------------------------------------------
# DAG Generating Functions


def dag_avg_deg(p, k, w_min=1, w_max=1, return_ordering=False, random_state=None, debug=False):
    """Generate an Erdos-Renyi graph with p nodes and average degree k,
    and orient edges according to a random ordering. Sample the edge
    weights from a uniform distribution.

    Parameters
    ----------
    p : int
        The number of nodes in the graph.
    k : float
        The desired average degree.
    w_min : float, optional
        The lower bound on the sampled weights. Defaults to 1.
    w_max : float, optional
        The upper bound on the sampled weights. Defaults to 1.
    return_ordering: bool, optional
        If the topological ordering used to orient the edges should be
        returned.
    random_state : int,optional
        To set the random state for reproducibility.
    debug : bool, optional
        If debug traces should be printed.

    Returns
    -------
    W : numpy.ndarray
       The connectivity (weights) matrix of the generated DAG.
    ordering : numpy.ndarray, optional
       If return_ordering = True, a topological ordering of the graph.

    Example
    -------

    >>> from sempler.generators import dag_avg_deg
    >>> dag_avg_deg(5, 2, random_state = 42)
    array([[0., 0., 1., 1., 0.],
           [0., 0., 1., 0., 0.],
           [0., 0., 0., 0., 0.],
           [0., 0., 1., 0., 1.],
           [0., 0., 0., 0., 0.]])

    Optionally, the ordering used to orient the edges can be returned

    >>> dag_avg_deg(5, 2, return_ordering = True, random_state = 42)
    (array([[0., 0., 1., 1., 0.],
           [0., 0., 1., 0., 0.],
           [0., 0., 0., 0., 0.],
           [0., 0., 1., 0., 1.],
           [0., 0., 0., 0., 0.]]), array([0, 3, 1, 4, 2]))


    """
    rng = np.random.default_rng(random_state)
    # Generate adjacency matrix as if top. ordering is 1..p
    prob = k / (p - 1)
    print("p = %d, k = %0.2f, P = %0.4f" % (p, k, prob)) if debug else None
    A = rng.uniform(size=(p, p))
    A = (A <= prob).astype(float)
    A = np.triu(A, k=1)
    weights = rng.uniform(w_min, w_max, size=A.shape)
    W = A * weights

    # Permute rows/columns according to random topological ordering
    permutation = rng.permutation(p)
    # Note the actual topological ordering is the "conjugate" of permutation eg. [3,1,2] -> [2,3,1]
    print("avg degree = %0.2f" % (np.sum(A) * 2 / len(A))) if debug else None
    if return_ordering:
        return (W[permutation, :][:, permutation], np.argsort(permutation))
    else:
        return W[permutation, :][:, permutation]


def dag_full(p, w_min=1, w_max=1, return_ordering=False, random_state=None):
    """Create a fully connected DAG, sampling the weights from a uniform
    distribution.

    Parameters
    ----------
    p : int
        The number of nodes in the graph.
    w_min : float, optional
        The lower bound on the sampled weights. Defaults to 1.
    w_max : float, optional
        The upper bound on the sampled weights. Defaults to 1.
    return_ordering: bool, optional
        If the topological ordering used to orient the edges should be
        returned.
    random_state : int,optional
        To set the random state for reproducibility.
    debug : bool, optional
        If debug traces should be printed.

    Returns
    -------
    W : numpy.ndarray
       The connectivity (weights) matrix of the generated DAG.
    ordering : numpy.ndarray, optional
       If return_ordering = True, a topological ordering of the graph.

    Example
    -------

    >>> from sempler.generators import dag_full
    >>> dag_full(4, random_state = 42)
    array([[0., 0., 1., 1.],
           [1., 0., 1., 1.],
           [0., 0., 0., 0.],
           [0., 0., 1., 0.]])


    Optionally, the ordering used to orient the edges can be returned

    >>> dag_full(4, return_ordering = True, random_state = 42)
    (array([[0., 0., 1., 1.],
           [1., 0., 1., 1.],
           [0., 0., 0., 0.],
           [0., 0., 1., 0.]]), array([1, 0, 3, 2]))



    """
    rng = np.random.default_rng(random_state)
    # Build a triangular matrix
    A = np.triu(np.ones((p, p)), k=1)
    weights = rng.uniform(w_min, w_max, size=A.shape)
    W = A * weights
    # Permute rows/columns according to random topological ordering
    permutation = rng.permutation(p)
    # Note the actual topological ordering is the "conjugate" of permutation eg. [3,1,2] -> [2,3,1]
    if return_ordering:
        return (W[permutation, :][:, permutation], np.argsort(permutation))
    else:
        return W[permutation, :][:, permutation]


def intervention_targets(p, K, size, replace=True, random_state=None):
    """Sample a set of intervention targets.

    Parameters
    ----------
    p : int
        The number of variables, i.e. targets will be sampled from
        `[0,p-1]`.
    K : int
        The total number of interventions.
    size : int or tuple
        Specifies the size of each intervention, i.e. the number of
        targets / intervention. If a two-element tuple, the number of
        targets is sampled uniformly at random from `[size[0],
        size[1]]`.
    replace : bool, default=True
        Wether the intervention targets should be sampled with
        replacement, i.e. if repeated targets are allowed across
        environments.
    random_state : int or None
        To set the random state for reproducibility.

    Returns
    -------
    interventions : list of list of int
        The sampled intervention targets.

    Raises
    ------
    ValueError :
        If the size of each intervention (i.e. number of targets) is
        larger than the actual number of variables, or if the tuple
        passed as size does not have length 2.

    Examples
    --------

    Generating a set of single-variable interventions:

    >>> from sempler.generators import intervention_targets
    >>> intervention_targets(10, 5, 1, random_state=42)
    [[0], [7], [6], [4], [4]]

    Without replacement:

    >>> intervention_targets(10, 5, 1, replace=False, random_state=42)
    [[0], [7], [6], [4], [3]]

    Generating a set of interventions with random number of targets:

    >>> intervention_targets(10, 5, (1,3), random_state=42)
    [[8], [2, 6, 0], [8, 7], [6, 7], [8, 1]]

    Without replacement:

    >>> intervention_targets(10, 5, (1,2), replace=False, random_state=42)
    [[8], [6, 0], [1, 4], [7], [9]]

    An exception is raised if `size > p`:

    >>> intervention_targets(4, 5, 5)
    Traceback (most recent call last):
      ...
    ValueError: The (max.) intervention size cannot be larger than the number of variables.

    Or if `size` is a tuple with size different than two:

    >>> intervention_targets(4, 5, (0,1,2))
    Traceback (most recent call last):
      ...
    ValueError: The intervention size must be a positive integer or two-element tuple.

    If sampling targets without replacement, the maximum intervention
    size and number of interventions must be set accordingly,
    i.e. `max_size x K <= p`. Otherwise an exception is raised:

    >>> intervention_targets(10, 5, (0,3), replace=False)
    Traceback (most recent call last):
      ...
    ValueError: Cannot sample targets without replacement for the given intervention size and number of interventions.


    """
    rng = np.random.default_rng(random_state)
    # Build intervention sizes
    if isinstance(size, tuple) and len(size) == 2:
        min_size, max_size = size
        sizes = rng.integers(size[0], size[1] + 1, K)
    elif isinstance(size, tuple):
        raise ValueError("The intervention size must be a positive integer or two-element tuple.")
    else:
        max_size = size
        sizes = [size] * K
    # If sampling without
    if not replace:
        if max_size * K > p:
            raise ValueError(
                "Cannot sample targets without replacement for the given intervention size and number of interventions.")
    # Check max size condition
    if max_size > p:
        raise ValueError(
            "The (max.) intervention size cannot be larger than the number of variables.")
    # Sample the targets
    if replace:
        interventions = []
        targets = list(range(p))
        for i, k in enumerate(range(K)):
            intervention = list(rng.choice(targets, size=sizes[i], replace=False))
            interventions.append(intervention)
    else:
        interventions = []
        remaining_targets = set(range(p))
        for i, k in enumerate(range(K)):
            intervention = list(rng.choice(list(remaining_targets), size=sizes[i], replace=False))
            remaining_targets -= set(intervention)
            interventions.append(intervention)
    return interventions


# To run the method's doctests
if __name__ == '__main__':
    import doctest
    doctest.testmod(verbose=True)



# Copyright 2021 Juan L Gamella

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Main module containing the SCM classes (sempler.ANM and
sempler.LGANM), the sempler.NormalDistribution class and additional
functions to generate random graphs.
"""

import numpy as np
from copy import deepcopy
import sempler.utils as utils
import sempler.functions as functions

# ---------------------------------------------------------------------
# ANM class


class ANM:
    """Class to represent a general (acyclic) additive noise model.

    Parameters
    ----------
    A : numpy.ndarray
        The `p x p` adjacency matrix specifying the functional
        dependencies, where `p` is the number of variables and `A[i,j]
        != 0` if `i` appears in the assignment of `j` (i.e. `i -> j`).
    assignments : list of (function or None)
        A list of p functions representing the functional assignments
        of each variable. Each function must take a vector of
        observations where each column corresponds to a parent as
        specified by the adjacency matrix `A`, or be `None` or
        `sempler.functions.null` if the variable has no parents.
    noise_distributions : list of function
        A list of `p` functions that generate samples from each
        variable's noise distribution (see sempler.noise for
        details).

    Raises
    ------
    ValueError
        If the given adjacency does not correspond to a DAG.

    Example
    -------

    Constructing an ANM.

    >>> import sempler
    >>> import sempler.noise as noise
    >>> import numpy as np

    (1) Define the connectivity matrix:

    >>> A = np.array([[0, 0, 0, 1, 0],
    ...               [0, 0, 1, 0, 0],
    ...               [0, 0, 0, 1, 0],
    ...               [0, 0, 0, 0, 1],
    ...               [0, 0, 0, 0, 0]])

    (2) Define the noise distributions (see sempler.noise):

    >>> noise_distributions = [noise.normal(0,1)] * 5

    (3) Define the variable assignments:

    >>> assignments = [None, None, np.sin, lambda x: np.exp(x[:,0]) + 2*x[:,1], lambda x: 2*x]

    Putting it all together:

    >>> anm = sempler.ANM(A, assignments, noise_distributions)

    An exception is raised if the adjacency matrix does not belong to a DAG:

    >>> A[4,0] = 1
    >>> sempler.LGANM(A, (0,0), (1,1))
    Traceback (most recent call last):
      ...
    ValueError: The given graph is not a DAG.

    Attributes
    ----------
    A : numpy.ndarray
        The adjacency matrix specifying the functional dependencies.
    p : int
        The number of variables (size) of the SCM.
    assignments : list of function
        The assignment functions of the variables.
    noise_distributions : list of function
        A list of functions representing the noise term distribution
        of each variable (see sempler.noise).

    """

    def __init__(self, A, assignments, noise_distributions):
        self.ordering = utils.topological_ordering(A)
        self.p = len(A)
        self.A = deepcopy(A)
        self.assignments = [functions.null if fun is None else deepcopy(fun) for fun in assignments]
        self.noise_distributions = deepcopy(noise_distributions)

    def sample(self, n, do_interventions={}, shift_interventions={}, noise_interventions={}, random_state=None):
        """Generates i.i.d. observations from the ANM, under the given do, shift or
        noise interventions. If no interventions are given, sample from the observational
        distribution.

        Parameters
        ----------
        n : int
            The size of the sample (i.e. number of observations).
        do_interventions : dict, optional
            A dictionary where keys correspond to the intervened
            variables, and the values are the distribution functions
            (see sempler.noise) to generate samples for each
            intervened variable.
        shift_interventions : dict, optional
            A dictionary where keys correspond to the intervened
            variables, and the values are the distribution functions
            (see sempler.noise) to generate noise samples which are
            added to the intervened variables.
        noise_interventions : dict, optional
            A dictionary where keys correspond to the intervened
            variables, and the values are the distribution functions
            (see sempler.noise) of the new noise.
        random_state: int, optional
            To set the random state for reproducibility,
            i.e. successive calls will return the same sample.

        Returns
        -------
        numpy.ndarray
            An array containing the sample, with each column
            corresponding to a variable.

        Examples
        --------

        Sampling from the observational setting:

        >>> samples = anm.sample(100)

        Sampling under a shift intervention on variable 1:

        >>> import sempler.noise as noise
        >>> samples = anm.sample(100, shift_interventions = {1: noise.normal(0,1)})

        Sampling under a noise intervention on variable 0 and a do intervention on variable 2:

        >>> samples = anm.sample(100,
        ...                      noise_interventions = {0: noise.normal()},
        ...                      do_interventions = {2 : noise.uniform()})

        """
        # Set random state (if requested)
        np.random.seed(random_state) if random_state is not None else None
        # Sample according to a topological ordering of the connectivity matrix
        X = np.zeros((n, self.p))
        for i in self.ordering:
            # If i is do intervened, sample from the corresponding
            # interventional distribution
            if i in do_interventions:
                X[:, i] = do_interventions[i](n)
            # Otherwise maintain dependence on parents
            else:
                assignment = np.transpose(self.assignments[i](X[:, self.A[:, i] != 0]))
                # Shift-intervention: add noise from given distribution
                if i in shift_interventions:
                    noise = self.noise_distributions[i](n) + shift_interventions[i](n)
                # Noise-intervention: sample noise from given distribution
                elif i in noise_interventions:
                    noise = noise_interventions[i](n)
                # No intervention: sample noise from original distribution
                else:
                    noise = self.noise_distributions[i](n)
                X[:, i] = assignment + noise
        return X


# To run the ANM.sample doctests
if __name__ == '__main__':
    import doctest
    import sempler.noise
    # Build ANM
    A = np.array([[0, 0, 0, 1, 0],
                  [0, 0, 1, 0, 0],
                  [0, 0, 0, 1, 0],
                  [0, 0, 0, 0, 1],
                  [0, 0, 0, 0, 0]])
    noise_distributions = [sempler.noise.normal(0, 1)] * 5
    assignments = [None, None, np.sin, lambda x: np.exp(x[:, 0]) + 2 * x[:, 1], lambda x: 2 * x]
    anm = ANM(A, assignments, noise_distributions)
    doctest.testmod(extraglobs={'anm': anm}, verbose=True)



import setuptools

setuptools.setup(
    name='sempler',
    version='0.2.3',
    author='Juan L Gamella',
    author_email='juangamella@gmail.com',
    packages=['sempler', 'sempler.test'],
    scripts=[],
    url='https://github.com/juangamella/sempler',
    license='BSD 3-Clause License',
    description='Sample from general structural causal models (SCMs).',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    install_requires=['numpy>=1.15.0', 'matplotlib', 'networkx>=2.2']
)



# Copyright 2021 Juan L Gamella

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#---------------------------------------------------------------------
# Unit testing for module sampling.py

import unittest

import numpy as np
import networkx as nx

import sempler
import sempler.utils as utils

# Tests for the LGANM class
class LGANM_Tests(unittest.TestCase):
    def test_basic(self):
        # Test the initialization of an LGANM object
        p = 10
        W = sempler.generators.dag_avg_deg(p, p/4, 1, 1)
        sem = sempler.LGANM(W, (0,0), (1,1))
        self.assertTrue((sem.variances == np.ones(p)).all())
        self.assertTrue((sem.means == np.zeros(p)).all())
        self.assertTrue(np.sum((sem.W == 0).astype(float) + (sem.W == 1).astype(float)), p*p)

    def test_basic_1(self):
        # Test the initialization of an LGANM object
        p = 5
        W = sempler.generators.dag_avg_deg(p, p/4, 1, 1)
        sem = sempler.LGANM(W, (0,0), (1,1))
        self.assertTrue((sem.variances == np.ones(p)).all())
        self.assertTrue((sem.means == np.zeros(p)).all())
        sem = sempler.LGANM(W, np.zeros(p), np.ones(p))
        self.assertTrue((sem.variances == np.ones(p)).all())
        self.assertTrue((sem.means == np.zeros(p)).all())
        with self.assertRaises(Exception):
            sempler.LGANM(W, (0,0), (0,1,2,3,4))
        with self.assertRaises(Exception):
            sempler.LGANM(W, (0,0,0,0,0), (0,1))
        with self.assertRaises(Exception):
            sempler.LGANM(W, (0,1,2,3), (0,0,0))
        with self.assertRaises(Exception):
            sempler.LGANM(W, (0,1,2,3))

    def test_memory(self):
        # Test that all arguments are copied and not simply stored by
        # reference
        variances = np.array([1,2,3])
        means = np.array([3,4,5])
        W = np.array([[0,1,0],[0,0,1],[0,0,0]])
        sem = sempler.LGANM(W, means, variances)
        # Modify and compare
        variances[0] = 0
        means[2] = 1
        W[0,0] = 2
        self.assertFalse((W == sem.W).all())
        self.assertFalse((variances == sem.variances).all())
        self.assertFalse((means == sem.means).all())
        
    def test_means(self):
        # Test that means are set correctly
        p = 10
        W = sempler.generators.dag_avg_deg(p, p/4, 1, 1)
        means = np.arange(p)
        sem = sempler.LGANM(W, means, (0,1))
        self.assertTrue((sem.means == means).all())

    def test_sampling_args(self):
        variances = np.array([1,2,3])
        means = np.array([3,4,5])
        W = np.array([[0,1,1],[0,0,1],[0,0,0]])
        sem = sempler.LGANM(W, means, variances)
        self.assertEqual(np.ndarray, type(sem.sample(n=1)))
        self.assertEqual(np.ndarray, type(sem.sample(n=1, shift_interventions = {})))
        self.assertEqual(np.ndarray, type(sem.sample(n=1, do_interventions = {})))
        self.assertEqual(np.ndarray, type(sem.sample(n=1, shift_interventions = None)))
        self.assertEqual(np.ndarray, type(sem.sample(n=1, do_interventions = None)))
        self.assertEqual(sempler.NormalDistribution, type(sem.sample(n=1, population=True)))
        self.assertEqual(sempler.NormalDistribution, type(sem.sample(population=True)))
        
    def test_sampling_1(self):
        # Test sampling of a DAG with one variable
        np.random.seed(42)
        p = 1
        n = round(1e6)
        W = sempler.generators.dag_full(p)
        sem = sempler.LGANM(W, (0,0), (1,1))
        # Observational data
        truth = np.random.normal(0,1,size=(n,1))
        samples = sem.sample(n, shift_interventions = {})
        self.assertTrue(utils.same_normal(truth, samples, atol=1e-1))
        # Under do intervention
        truth = np.ones((n,1))
        samples = sem.sample(n, do_interventions = {0: 1})
        self.assertTrue((truth == samples).all())
        # Under noise intervention
        truth = np.random.normal(1,2,size=(n,1))
        samples = sem.sample(n, do_interventions = {0: (1, 4)})
        self.assertTrue(utils.same_normal(truth, samples, atol=1e-1))

    def test_sampling_2(self):
        # Test that the distribution of a 4 variable DAG with upper
        # triangular, all ones adj. matrix matches what we expect
        # using the path method
        p = 4
        n = round(1e6)
        W = np.triu(np.ones((p,p)),k=1)
        sem = sempler.LGANM(W, (0,0), (0.16,0.16))
        np.random.seed(42)
        noise = np.random.normal([0,0,0,0],[.4, .4, .4, .4], size=(n,4))
        truth = np.zeros((n,p))
        truth[:,0] = noise[:,0]
        truth[:,1] = noise[:,0] + noise[:,1]
        truth[:,2] = 2*noise[:,0] + noise[:,1] + noise[:,2]
        truth[:,3] = 4*noise[:,0] + 2*noise[:,1] + noise[:,2] + noise[:,3]
        samples = sem.sample(n)
        self.assertTrue(utils.same_normal(truth, samples))
        
    def test_interventions_1(self):
        # Test sampling and interventions on a custom DAG, comparing
        # with results obtained via the path method
        np.random.seed(42)
        p = 6
        n = round(1e6)
        W = np.array([[0, 1, 1, 0, 0, 0],
                      [0, 0, 0, 0, 1, 1],
                      [0, 0, 0, 1, 0, 0],
                      [0, 0, 0, 0, 1, 1],
                      [0, 0, 0, 0, 0, 1],
                      [0, 0, 0, 0, 0, 0]])
        sem = sempler.LGANM(W, (0,0), (0.16,0.16))

        # Test observational data
        M = np.array([[1, 0, 0, 0, 0, 0],
                      [1, 1, 0, 0, 0, 0],
                      [1, 0, 1, 0, 0, 0],
                      [1, 0, 1, 1, 0, 0],
                      [2, 1, 1, 1, 1, 0],
                      [4, 2, 2, 2, 1, 1]])
        noise = np.random.normal(np.zeros(p), np.ones(p)*0.4, size=(n,p))
        truth = noise @ M.T
        samples = sem.sample(n)
        self.assertTrue(utils.same_normal(truth, samples))
        
        # Test under do-interventions on X1
        noise = np.random.normal([2.1,0,0,0,0,0], [0,.4, .4, .4, .4, .4], size=(n,p))
        truth = noise @ M.T
        samples = sem.sample(n, do_interventions = {0: 2.1})
        self.assertTrue(utils.same_normal(truth, samples))
        
        # Test under do-intervention on X1 and noise interventions X2 and X5
        do_int = {0: 2, 1: (2, 0.25), 4: (1, 0.25)}
        noise = np.random.normal([2,2,0,0,1,0], [0,.5,.4,.4,.5,.4], size=(n,p))
        M = np.array([[1, 0, 0, 0, 0, 0],
                      [0, 1, 0, 0, 0, 0],
                      [1, 0, 1, 0, 0, 0],
                      [1, 0, 1, 1, 0, 0],
                      [0, 0, 0, 0, 1, 0],
                      [1, 1, 1, 1, 1, 1]])
        truth = noise @ M.T
        samples = sem.sample(n, do_interventions=do_int)
        self.assertTrue(utils.same_normal(truth, samples))

    def test_interventions_2(self):
        # Test that the means and variances of variables in the joint
        # distribution are what is expected via the path method
        W = np.array([[0, 1, 1],
                      [0, 0, 1],
                      [0, 0, 0]])
        n = round(1e6)
        variances = np.array([1,2,3])*0.1
        means = np.array([1,2,3])
        sem = sempler.LGANM(W, means, variances)
        np.random.seed(42)
        # Test observational data
        # Build truth
        noise = np.random.normal(means, variances**0.5, size=(n,3))
        truth = np.zeros_like(noise)
        truth[:,0] = noise[:,0]
        truth[:,1] = truth[:,0]*W[0,1] + noise[:,1]
        truth[:,2] = truth[:,0]*W[0,2] + truth[:,1]*W[1,2] + noise[:,2]
        samples = sem.sample(n)
        self.assertTrue(utils.same_normal(truth, samples))
        # Test that variances/means are as expected
        true_vars, true_means = np.zeros(3), np.zeros(3)
        true_vars[0] = variances[0]
        true_vars[1] = W[0,1]**2 * variances[0] + variances[1]
        true_vars[2] = (W[0,1]*W[1,2] + W[0,2])**2 * variances[0] + W[1,2]**2 * variances[1] + variances[2]
        true_means[0] = means[0]
        true_means[1] = W[0,1] * means[0] + means[1]
        true_means[2] = (W[0,1] * W[1,2] + W[0,2]) * means[0] + W[1,2] * means[1] + means[2]
        self.assertTrue(np.allclose(true_vars, np.var(samples, axis=0), atol=1e-2))
        self.assertTrue(np.allclose(true_means, np.mean(samples, axis=0), atol=1e-2))
        
        # Test under intervention on X1 <- N(0,0.1)
        variances = np.array([1., 1., 3.])*0.1
        means = np.array([1., 0., 3.])
        noise = np.random.normal(means, variances**0.5, size=(n,3))
        truth[:,0] = noise[:,0]
        truth[:,1] = noise[:,1]
        truth[:,2] = truth[:,0]*W[0,2] + truth[:,1]*W[1,2] + noise[:,2]
        samples = sem.sample(n, do_interventions = {1: (0,0.1)})
        self.assertTrue(utils.same_normal(truth, samples))
        # Test that variances/means are as expected
        true_vars, true_means = np.zeros(3), np.zeros(3)
        true_vars[0] = variances[0]
        true_vars[1] = variances[1]
        true_vars[2] = W[0,2]**2 * variances[0] + W[1,2]**2 * variances[1] + variances[2]
        true_means[0] = means[0]
        true_means[1] = means[1]
        true_means[2] = W[0,2] * means[0] + W[1,2] * means[1] + means[2]
        self.assertTrue(np.allclose(true_vars, np.var(samples, axis=0), atol=1e-2))
        self.assertTrue(np.allclose(true_means, np.mean(samples, axis=0), atol=1e-2))
        
        # Test under intervention on do(X0 = 0)
        variances = np.array([0., 2., 3.])*0.1
        means = np.array([0., 2., 3.])
        noise = np.random.normal(means, variances**0.5, size=(n,3))
        truth[:,0] = noise[:,0]
        truth[:,1] = truth[:,0]*W[0,1] + noise[:,1]
        truth[:,2] = truth[:,0]*W[0,2] + truth[:,1]*W[1,2] + noise[:,2]
        samples = sem.sample(n, do_interventions = {0: 0})
        self.assertTrue(utils.same_normal(truth, samples))
        # Test that variances/means are as expected
        true_vars, true_means = np.zeros(3), np.zeros(3)
        true_vars[0] = variances[0]
        true_vars[1] = W[0,1]**2 * variances[0] + variances[1]
        true_vars[2] = (W[0,1]*W[1,2] + W[0,2])**2 * variances[0] + W[1,2]**2 * variances[1] + variances[2]
        true_means[0] = means[0]
        true_means[1] = W[0,1] * means[0] + means[1]
        true_means[2] = (W[0,1] * W[1,2] + W[0,2]) * means[0] + W[1,2] * means[1] + means[2]
        self.assertTrue(np.allclose(true_vars, np.var(samples, axis=0), atol=1e-2))
        self.assertTrue(np.allclose(true_means, np.mean(samples, axis=0), atol=1e-2))

    def test_distribution(self):
        # Test "population" sampling
        W = np.array([[0,0,1,0],
                      [0,0,1,0],
                      [0,0,0,1],
                      [0,0,0,0]])
        # Build SEM with unit weights and standard normal noise
        # variables
        sem = sempler.LGANM(W, (0,0), (1,1))
        # Observational Distribution
        distribution = sem.sample(population=True)
        true_cov = np.array([[1,0,1,1],
                             [0,1,1,1],
                             [1,1,3,3],
                             [1,1,3,4]])
        self.assertTrue((distribution.mean==np.zeros(4)).all())
        self.assertTrue((distribution.covariance==true_cov).all())
        # Do intervention on X1 <- 0
        distribution = sem.sample(population=True, do_interventions = {0: 1})
        true_cov = np.array([[0,0,0,0],
                             [0,1,1,1],
                             [0,1,2,2],
                             [0,1,2,3]])
        self.assertTrue((distribution.mean==np.array([1,0,1,1])).all())
        self.assertTrue((distribution.covariance==true_cov).all())
        # Noise interventions on X1 <- N(0,2), X2 <- N(1,2)
        interventions = {0: (0,2), 1: (1,2)}
        distribution = sem.sample(population=True, do_interventions=interventions)
        true_cov = np.array([[2,0,2,2],
                             [0,2,2,2],
                             [2,2,5,5],
                             [2,2,5,6]])
        self.assertTrue((distribution.mean==np.array([0,1,1,1])).all())
        self.assertTrue((distribution.covariance==true_cov).all())

    def test_shift_noise_interventions_1(self):
        # Test sampling and interventions on a custom DAG, comparing
        # with results obtained via the path method
        np.random.seed(42)
        p = 6
        n = round(1e6)
        W = np.array([[0, 1, 1, 0, 0, 0],
                      [0, 0, 0, 0, 1, 1],
                      [0, 0, 0, 1, 0, 0],
                      [0, 0, 0, 0, 1, 1],
                      [0, 0, 0, 0, 0, 1],
                      [0, 0, 0, 0, 0, 0]])
        sem = sempler.LGANM(W, (0,0), (0.16,0.16))

        # Test observational data
        M = np.array([[1, 0, 0, 0, 0, 0],
                      [1, 1, 0, 0, 0, 0],
                      [1, 0, 1, 0, 0, 0],
                      [1, 0, 1, 1, 0, 0],
                      [2, 1, 1, 1, 1, 0],
                      [4, 2, 2, 2, 1, 1]])
        noise = np.random.normal(np.zeros(p), np.ones(p)*0.4, size=(n,p))
        truth = noise @ M.T
        samples = sem.sample(n)
        self.assertTrue(utils.same_normal(truth, samples))

        # Test shift intervention on X4
        noise = np.random.normal([0,0,0,0,0,0], [.4, .4, .4, .4, .5, .4], size=(n,p))
        truth = noise @ M.T
        samples = sem.sample(n, shift_interventions = {4: (0,0.09)})
        self.assertTrue(utils.same_normal(truth, samples))

        # Test under noise intervention on X4
        noise = np.random.normal([0,0,0,0,0,0], [.4, .4, .4, .4, 1, .4], size=(n,p))
        truth = noise @ M.T
        samples = sem.sample(n, noise_interventions = {4: (0,1)})
        self.assertTrue(utils.same_normal(truth, samples))
        
        # Test noiseless shift intervention on X2
        noise = np.random.normal([0,0,2,0,0,0], [.4, .4, .4, .4, .4, .4], size=(n,p))
        truth = noise @ M.T
        samples = sem.sample(n, shift_interventions = {2: 2})
        self.assertTrue(utils.same_normal(truth, samples))
        
        # Test that do-interventions on X0 override shift interventions on X0
        noise = np.random.normal([2.1,0,0,0,0,0], [0,.4, .4, .4, .4, .4], size=(n,p))
        truth = noise @ M.T
        samples = sem.sample(n, do_interventions = {0: 2.1}, shift_interventions = {0: (1,2)})
        self.assertTrue(utils.same_normal(truth, samples))
        
        # Test under shift-intervention on X0 and do interventions X1 and X4
        shift_int = {0: (0,0.2)}
        do_int = {1: (2, 0.25), 4: (1, 0.25)}
        noise = np.random.normal([0,2,0,0,1,0], [0.6,.5,.4,.4,.5,.4], size=(n,p))
        M = np.array([[1, 0, 0, 0, 0, 0],
                      [0, 1, 0, 0, 0, 0],
                      [1, 0, 1, 0, 0, 0],
                      [1, 0, 1, 1, 0, 0],
                      [0, 0, 0, 0, 1, 0],
                      [1, 1, 1, 1, 1, 1]])
        truth = noise @ M.T
        samples = sem.sample(n, do_interventions=do_int, shift_interventions=shift_int)
        self.assertTrue(utils.same_normal(truth, samples))    





# Copyright 2021 Juan L Gamella

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#---------------------------------------------------------------------
# Unit testing for module sampling.py

import unittest
import numpy as np

import sempler
import sempler.noise
import sempler.utils
import sempler.generators

# Tests that the interface behaves as expected
class api_Tests(unittest.TestCase):
        
    def test_constructor(self):
        functions = [None, lambda x: x, lambda x: x]
        noises = [sempler.noise.normal()] * 3
        # Should raise exception for graph with cycle
        A = np.array([[0,1,0],
                      [0,0,1],
                      [1,0,0]])
        try:
            sempler.ANM(A, functions, noises)
            self.fail("Should have raised an exception")
        except ValueError as e:
            print("OK;",e)
        # Should raise exception for PDAG
        A = np.array([[0,1,0],
                      [1,0,1],
                      [0,0,0]])
        try:
            sempler.ANM(A, functions, noises)
            self.fail("Should have raised an exception")
        except ValueError as e:
            print("OK;",e)
        # Should work
        A = np.array([[0,1,0],
                      [0,0,1],
                      [0,0,0]])
        sempler.ANM(A, [None, None, None], noises)
        # Test attributes
        anm = sempler.ANM(A, functions, noises)
        anm.A
        anm.p
        anm.assignments
        anm.noise_distributions
        # Test sampling
        # Should fail if n not specified
        try:
            anm.sample()
            self.fail("Should have raised an exception")
        except Exception as e:
            print("OK;",e)
        anm.sample(0)
        anm.sample(10)
        # Test empty interventions
        anm.sample(10, do_interventions = {})
        anm.sample(10, noise_interventions = {})
        anm.sample(10, shift_interventions = {})
        # Test non-empty interventions
        anm.sample(10, do_interventions = {0: sempler.noise.uniform()})
        anm.sample(10, noise_interventions = {0: sempler.noise.uniform()})
        anm.sample(10, shift_interventions = {0: sempler.noise.uniform()})

class sampling_Tests(unittest.TestCase):

    def test_gaussian_sampling(self):
        # Test 100 interventions
        K = 50
        W = np.array([[0, 0, 0, 0.2, 0],
                      [0, 0, 0.4, 0, 0],
                      [0, 0, 0, 0.3, 0],
                      [0, 0, 0, 0, 0.5],
                      [0, 0, 0, 0, 0  ]])
        lganm = sempler.LGANM(W, (1,2), (1,2))
        noise_distributions = [sempler.noise.normal(m,v) for (m,v) in zip(lganm.means, lganm.variances)]
        assignments = [None, None, lambda x: .4 * x, lambda x: .2*x[:,0] + .3 * x[:,1], lambda x: .5*x]
        anm = sempler.ANM(W, assignments, noise_distributions)
        interventions = sempler.generators.intervention_targets(lganm.p, K, (0,3))
        for targets in interventions:
            print(targets)
            means, variances = np.random.uniform(0,5,len(targets)), np.random.uniform(2,3,len(targets))
            interventions_lganm = dict((t, (m, v)) for (t,m,v) in zip(targets, means, variances))
            interventions_anm = dict((t, sempler.noise.normal(m,v)) for (t,m,v) in zip(targets, means, variances))
            # Sample each SCMs
            # TODO: Combine different interventions in one
            n = round(1e6)
            if len(targets) <= 1:
                samples_anm = anm.sample(n, do_interventions = interventions_anm)
                samples_lganm = lganm.sample(n, do_interventions = interventions_lganm)
            elif len(targets) == 2:
                samples_anm = anm.sample(n, shift_interventions = interventions_anm)
                samples_lganm = lganm.sample(n, shift_interventions = interventions_lganm)
            elif len(targets) == 3:
                samples_anm = anm.sample(n, noise_interventions = interventions_anm)
                samples_lganm = lganm.sample(n, noise_interventions = interventions_lganm)
            # Check that the distribution is the same
            self.assertTrue(sempler.utils.same_normal(samples_anm, samples_lganm, debug=False))



# Copyright 2021 Juan L Gamella

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#---------------------------------------------------------------------
# Unit testing for the NormalDistribution class

import unittest
import numpy as np

import sempler

from sempler.utils import sampling_matrix, all_but

# Tested functions
from sempler import NormalDistribution

class NormalDistributionTests(unittest.TestCase):
    def test_initialization(self):
        # Test initialization and that arguments are copied (not
        # stored by reference)
        covariance = np.array([[11,12,13],
                               [21,22,23],
                               [31,32,33]])
        mean = np.array([1,2,3])
        distribution = NormalDistribution(mean, covariance)
        self.assertTrue((mean == distribution.mean).all())
        self.assertTrue((covariance == distribution.covariance).all())
        mean[2] = 0
        covariance[:,2] = 0
        self.assertTrue(not (mean == distribution.mean).all())
        self.assertTrue(not (covariance == distribution.covariance).all())

    def test_equal(self):
        covariance = np.array([[11,12,13],
                               [21,22,23],
                               [31,32,33]])
        mean = np.array([1,2,3])
        dist1 = NormalDistribution(mean, covariance)
        dist2 = NormalDistribution(mean, covariance)
        self.assertTrue(dist1.equal(dist2))
        self.assertTrue(dist2.equal(dist1))
        tol = 1e-7
        dist3 = NormalDistribution(mean + tol/2, covariance + tol/2)
        self.assertTrue(dist1.equal(dist3))
        self.assertTrue(dist2.equal(dist3))
        self.assertTrue(dist3.equal(dist1))
        self.assertTrue(dist3.equal(dist2))
        
    def test_parameters(self):
        covariance = np.array([[11,12,13],
                               [21,22,23],
                               [31,32,33]])
        mean = np.array([1,2,3])
        dist = NormalDistribution(mean, covariance)
        # Marginalization
        dist_a = dist.marginal(0)
        dist_b = dist.marginal([0])
        self.assertTrue(dist_a.equal(dist_b))
        # Conditioning 1
        dist_a = dist.conditional(0, 1, 0)
        dist_b = dist.conditional([0], [1], [0])
        self.assertTrue(dist_a.equal(dist_b))
        # Conditioning 2
        dist_a = dist.conditional(0, [], [])
        try:
            dist.conditional(0, [], [1, 2])
            self.fail("Exception should have been raised")
        except ValueError as e:
            print("OK:",e)
        dist_c = dist.marginal(0)
        self.assertTrue(dist_a.equal(dist_c))
        
    def test_marginalization(self):
        # Test marginalization logic
        covariance = np.array([[11,12,13],
                               [21,22,23],
                               [31,32,33]])
        mean = np.array([1,2,3])
        distribution = NormalDistribution(mean, covariance)
        # Marginalize X1
        marginal = distribution.marginal([0])
        self.assertTrue((marginal.mean == np.array([1])).all())
        self.assertTrue((marginal.covariance == np.array([11])).all())
        # Marginalize X1 and X3
        marginal = distribution.marginal([0,2])
        self.assertTrue((marginal.mean == np.array([1, 3])).all())
        self.assertTrue((marginal.covariance == np.array([[11, 13], [31, 33]])).all())

    def test_regression_arguments(self):
        p = 5
        covariance = np.eye(p)
        mean = np.random.uniform(size=p)
        joint = NormalDistribution(mean, covariance)
        # Test can call with regressors as np.array (see issue #3)
        joint.regress(0, [1,2])
        joint.regress(0, np.array([1,2]))
        
    def test_conditioning_1(self):
        # Test conditioning
        covariance = np.eye(3)
        mean = np.zeros(3)
        distribution = NormalDistribution(mean, covariance)
        # Condition X1 on X2=1 should yield the marginal of X1 (as they are independent)
        conditional = distribution.conditional([0], [1], np.array([1]))
        self.assertTrue((conditional.mean == np.array([0])).all())
        self.assertTrue((conditional.covariance == np.array([1])).all())
        self.assertTrue((conditional.mean == distribution.marginal([0]).mean).all())
        self.assertTrue((conditional.covariance == distribution.marginal([0]).covariance).all())
        # Conditioning X1, X2 on X3 = 1 should yield marginal of X1, X2
        conditional = distribution.conditional([0,1], [2], 0)
        self.assertTrue((conditional.mean == np.array([0, 0])).all())
        self.assertTrue((conditional.covariance == np.eye(2)).all())
        self.assertTrue((conditional.mean == distribution.marginal([0,1]).mean).all())
        self.assertTrue((conditional.covariance == distribution.marginal([0,1]).covariance).all())
        # Conditioning X1 on X2=2, X3 = 1 should yield marginal of X1, X2
        conditional = distribution.conditional([0], [1,2], np.array([2,1]))
        self.assertTrue((conditional.mean == np.array([0])).all())
        self.assertTrue((conditional.covariance == np.eye(1)).all())
        self.assertTrue((conditional.mean == distribution.marginal([0]).mean).all())
        self.assertTrue((conditional.covariance == distribution.marginal([0]).covariance).all())

    def test_conditioning_2(self):
        covariance = np.array([[1,2,3], [2,4,6], [3,6,9]]) # rank 1
        mean = np.array([1,2,3])
        dist = NormalDistribution(mean, covariance)
        # Conditioning X1 on X2 = 1 should give mean = 1/2 and variance = 0
        self.assertTrue((dist.conditional([0], [1], 1).mean == np.array([0.5])).all())
        self.assertTrue((dist.conditional([0], [1], 1).covariance == np.array([[0]])).all())
        # Conditioning X1 on X2 = 0 should give mean = 0 and variance = 0
        self.assertTrue((dist.conditional([0], [1], 0).mean == np.array([0])).all())
        self.assertTrue((dist.conditional([0], [1], 0.5).covariance == np.array([[0]])).all())

    def test_conditioning_3(self):
        covariance = np.array([[1, 0, 1, 1],
                               [0, 1, 1, 1],
                               [1, 1, 3, 3],
                               [1, 1, 3, 4]])
        mean = np.array([0,0,0,0])
        dist = NormalDistribution(mean, covariance)
        # Conditioning X3 on X1 = 1, X2 = 2 should give mean 3 and variance 1
        conditional = dist.conditional([2], [0,1], np.array([1,2]))
        self.assertTrue((conditional.mean == np.array([3])).all())
        self.assertTrue((conditional.covariance == np.array([[1]])).all())
        # Conditioning X4 on X1 = 1, X2 = 2 should give mean 3 and variance 2
        conditional = dist.conditional([3], [0,1], np.array([1,2]))
        self.assertTrue((conditional.mean == np.array([3])).all())
        self.assertTrue((conditional.covariance == np.array([[2]])).all())
        # Conditioning X3, X4 on X1 = 1, X2 = 2
        conditional = dist.conditional([2, 3], [0,1], np.array([1,2]))
        self.assertTrue((conditional.mean == np.array([3,3])).all())
        self.assertTrue((conditional.covariance == np.array([[1,1], [1,2]])).all())
        # Conditioning X3 on X1 = 1, X4 = 1 should give mean 1 and variance 3-7/3
        conditional = dist.conditional([2], [0,3], np.array([1,1]))
        self.assertTrue(np.allclose(conditional.mean, np.array([1])))
        self.assertTrue(np.allclose(conditional.covariance, np.array([[3-7/3]])))
        
    def test_sampling(self):
        # Test sampling of the distribution
        np.random.seed(42)
        n = round(1e6)
        covariance = 1e-2 * np.array([[1, 0, 1, 1],
                                      [0, 1, 1, 1],
                                      [1, 1, 3, 3],
                                      [1, 1, 3, 4]])
        mean = np.array([0,0,0,0])
        dist = NormalDistribution(mean, covariance)
        samples = dist.sample(n)
        self.assertTrue(np.allclose(np.mean(samples, axis=0), mean, atol=1e-1))
        self.assertTrue(np.allclose(np.cov(samples, rowvar=False), covariance,  atol=1e-1))

    def run_mse_tests(self, joint, tests):
        for test in tests:
            (y, Xs, true_mse) = test
            result = joint.mse(y, Xs)
            if not np.allclose(true_mse, result):
                print(test)
                self.fail("%0.16f != %0.16f" % (true_mse, result))

    def mse_properties_test(self, W, parents, markov_blankets, sample_weights=True):
        # Given a weight matrix, parents of each node and markov
        # blankets of each node, test basic properties of the MSE
        # PRE: It is assumed that [1,...,p] is the causal ordering
        p = len(W)
        np.random.seed(42)
        if sample_weights:
            W = W * (np.random.uniform(size=W.shape) + 1) # avoid 0 weights (causal minimality)
        A = sampling_matrix(W)
        covariance = A @ A.T
        mean = np.random.uniform(size=p)
        joint = NormalDistribution(mean, covariance)
        # Tests
        tests = []
        # When no regressors, the MSE is just the marginal variance
        tests += [(i, [], covariance[i,i]) for i in range(p)]
        # Regressing a variable on itself should give 0 MSE
        tests += [(i, [i], 0) for i in range(p)]
        # Regressing on all variables (including itself) should also
        # give 0 MSE
        tests += [(i, range(p), 0) for i in range (p)]
        # Regressing on the parents should yield MSE equal to
        # the variance of the noise variable (1 in this case)
        tests += [(i, pa, 1) for i,pa in enumerate(parents)]
        # Regressing on additional variables to the Markov Blanket
        # should give the same result
        tests += [(i, all_but(i, p), joint.mse(i, mb)) for i,mb in enumerate(markov_blankets)]
        self.run_mse_tests(joint, tests)
        # Changing the mean of the noise variables should not change
        # the MSE
        joint.mean = np.random.uniform(size=p)
        self.run_mse_tests(joint, tests)
        
    def regression_properties_test(self, W, parents, markov_blankets):
        # Test basic properties of the regression coefficients.
        # PRE: Assumed [1..p] is a valid causal ordering
        p = len(W)
        np.random.seed(42)
        W = W * (np.random.uniform(size=W.shape) + 1) # avoid 0 weights (causal minimality)
        A = sampling_matrix(W)
        covariance = A @ A.T
        mean = np.random.uniform(size=p)
        joint = NormalDistribution(mean, covariance)
        # Tests
        # Regressing on the parents should return their weights as coefficients
        for i,pa in enumerate(parents):
            (coefs, intercept) = joint.regress(i, pa)
            if not np.allclose(W[pa, i], coefs[pa]):
                self.fail("%s ~ %s. Expected: %s Got: %s" % (i, pa, W[pa, i], coefs[pa]))
        # Regressing on all variables should yield weight 1 on the
        # target and 0 intercept
        for i in range(p):
            (coefs, intercept) = joint.regress(i, range(p))
            truth = np.zeros(p)
            truth[i] = 1
            if not np.allclose(truth, coefs):
                self.fail("%s ~ %s. Expected: %s Got: %s" % (i, pa, truth, coefs))
            if not np.allclose(0, intercept):
                self.fail("%s ~ %s. Expected: %s Got: %s" % (i, pa, 0, intercept))
        # Regressing on all variables except target should yield Markov blanket
        for i,mb in enumerate(markov_blankets):
            mb.sort()
            (coefs, intercept) = joint.regress(i, all_but(i, p))
            non_mb = [i for i in range(p) if i not in mb]
            # Check vars. outside MB have 0 coefficient
            truth = np.zeros(len(non_mb))
            if not np.allclose(truth, coefs[non_mb]):
                self.fail("%s ~ %s. Expected: %s Got: %s" % (i, mb, truth, coefs[non_mb]))
            # Check that vars. inside MB have non-zero coefficient (we
            # assumed causal minimality above)
            tol = 1e-10
            if not (np.abs(coefs[mb]) > tol).all():
                self.fail("%s ~ %s. Expected all non-zero. Got: %s" % (i, mb, coefs[mb]))

    def test_mse_1(self):
        # Test a set of regressions
        covariance = np.array([[1, 0, 1, 1],
                               [0, 1, 1, 1],
                               [1, 1, 3, 3],
                               [1, 1, 3, 4]])
        mean = np.array([0,0,0,0])
        joint = NormalDistribution(mean, covariance)
        tests = [
            # Regressing X0
            (0, [1], 1), # indep. var
            (0, [2], 2/3), # children
            (0, [1,2], 0.5), # markov blanket
            (0, [2,1], 0.5), # markov blanket (test order does not matter)
            # Regressing X1
            (1, [0], 1), # indep. var
            (1, [2], 2/3), # children
            (1, [0,2], 0.5), # markov blanket
            (1, [2,0], 0.5), # markov blanket (test order does not matter)
            # Regressing X2
            (2, [0], 2), # parent
            (2, [1], 2), # parent
            (2, [0,1], 1), # parents
            (2, [0,1,3], 0.5), # markov blanket
            (2, [3], 0.75), # child
            # Regressing X3
            (3, [0], 3), # X0
            (3, [0], 3), # X1
            (3, [0,1], 2), # X0 and X1
            (3, [1,0], 2), # X1 and X0 (test order does not matter)
            (3, [2,1], 1), # markov blanket + d-separated
            (3, [1,2], 1), # markov blanket + d-separated (test order does not matter)
            (3, [2,0], 1), # markov blanket + d-separated
            (3, [2], 1)] # markov blanket
        for test in tests:
            (y, Xs, true_mse) = test
            self.assertEqual(true_mse, joint.mse(y, Xs))
        # Changing the mean of the noise variables should not change
        # the results
        joint.mean = np.array([1,2,3,4])
        for test in tests:
            (y, Xs, true_mse) = test
            self.assertEqual(true_mse, joint.mse(y, Xs))
            
    def test_eg1(self):
        W = np.array([[0, 1, -1, 0, 0],
                      [0, 0, 0, 1, 0],
                      [0, 0, 0, 1, 0],
                      [0, 0, 0, 0, 1],
                      [0, 0, 0, 0, 0]])
        markov_blankets = [[1,2],
                           [0,3,2],
                           [0,3,1],
                           [1,2,4],
                           [3]]
        parents = [[],
                   [0],
                   [0],
                   [1,2],
                   [3]]
        self.mse_properties_test(W, parents, markov_blankets)
        self.mse_properties_test(W, parents, markov_blankets, sample_weights=False)
        self.regression_properties_test(W, parents, markov_blankets)
        
    def test_eg2(self):
        W = np.array([[0, 1, -1, 0, 0, 0],
                      [0, 0, 0, 1, 0, 0],
                      [0, 0, 0, 1, 0, 0],
                      [0, 0, 0, 0, 1, 0],
                      [0, 0, 0, 0, 0, 0],
                      [0, 0, 0, 0, 1, 0]])
        markov_blankets = [[1,2],
                           [0,3,2],
                           [0,3,1],
                           [1,2,4,5],
                           [3,5],
                           [3,4]]
        parents = [[],
                   [0],
                   [0],
                   [1,2],
                   [3,5],
                   []]
        self.mse_properties_test(W, parents, markov_blankets)
        self.mse_properties_test(W, parents, markov_blankets, sample_weights=False)
        self.regression_properties_test(W, parents, markov_blankets)

    def test_eg3(self):
        W = np.array([[0, 1, -1, 0, 0, 0, 0, 0],
                      [0, 0, 0, 1, 0, 0, 0, 0],
                      [0, 0, 0, 1, 0, 0, 0, 0],
                      [0, 0, 0, 0, 1, 0, 0, 0],
                      [0, 0, 0, 0, 0, 0, 0, 1],
                      [0, 0, 0, 0, 1, 0, 1, 0],
                      [0, 0, 0, 0, 0, 0, 0, 0],
                      [0, 0, 0, 0, 0, 0, 0, 0]])
        markov_blankets = [[1,2],
                           [0,3,2],
                           [0,3,1],
                           [1,2,4,5],
                           [3,5,7],
                           [3,4,6],
                           [5],
                           [4]]
        parents = [[],
                   [0],
                   [0],
                   [1,2],
                   [3,5],
                   [],
                   [5],
                   [4]]
        self.mse_properties_test(W, parents, markov_blankets)
        self.mse_properties_test(W, parents, markov_blankets, sample_weights=False)
        self.regression_properties_test(W, parents, markov_blankets)

    def test_eg4(self):
        W = np.array([[0,0,1,0],
                      [0,0,1,0],
                      [0,0,0,1],
                      [0,0,0,0]])
        markov_blankets = [[1,2],
                           [0,2],
                           [0,1,3],
                           [2]]
        parents = [[],
                   [],
                   [0,1],
                   [2]]
        self.mse_properties_test(W, parents, markov_blankets)
        self.mse_properties_test(W, parents, markov_blankets, sample_weights=False)
        self.regression_properties_test(W, parents, markov_blankets)
